#!/usr/bin/perl -wT

# Name:    Project Gutenberg search
# Author:  Manuel Arriaga (m.arriaga@ip.pt)
# Purpose: Searching the Project Gutenberg etexts for certain
#          words and/or phrases and retrieving the matching
#          sentences or paragraphs.
#
# Date:  2001/11/28
#
#
# Copyright 2001 Manuel Arriaga. 
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#	    
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#			    
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
					
#
#
# Configuration: One variable, $pg_root_dir, should be set,
#                otherwise the script will retrieve all files 
#                from the ibiblio.org mirror of the PG collection
#                each time it is run. See the 'Configuration' 
#                section below.
# Notes:         The HTML code which invokes this script should 
#                provide the values of the following form fields: 
#                author, title, words and type, the latter having 
#                one of two values: 'sentence' or 'paragraph'.
#                This code requires the modules LWP, CGI, IO::Scalar
#                and Archive::Zip.

use English;
use strict;

use LWP::Simple qw(get);
use CGI qw(header start_html end_html param h1 h3 h3 h3 h5 h6 p i var hr);

use IO::Scalar;
use Archive::Zip qw(:ERROR_CODES :CONSTANTS);

# --------------------------- Configuration ------------------------


# The only information you must provide is the path or URL of the
# PG root dir we will use. If this script is running on a PG mirror 
# site, then $pg_root_dir should be the _absolute_ path of the PG collection 
# root dir on this machine. If you aren't a mirror site of PG, then you 
# should provide the _complete_ URL of the closest (FTP/HTTP) PG mirror site,
# not forgetting to specify which protocol should be used. Just fill in the 
# correct information: (By default, the main PG mirror site is used so that 
# everything works "out of the box" on all computers.)

# Note: Since the script establishes a new connection for each file it
# retrieves, HTTP sites should be preferred.


my $pg_root_dir = 'http://ibiblio.org/pub/docs/books/gutenberg';


# ------------------------ End of configuration -------------------
#           [You don't need to change anything below this line.]



# Supporting subroutines ----------------------------------------------

# is_op returns 1 if the argument is a boolean operator (or a
# combination of two):

sub is_op
{
    my $token = shift();
    
    # As explained inside reformulate_query(), '&& !' and '|| !' are considered
    # operators:
    
    if ($token eq '||' || 
	$token eq '&&' || 
	$token eq '!'  ||
	$token eq '&& !' ||
	$token eq '|| !')
    {
	return 1;
    }
    else
    {
	return 0;
    }
}

# ----------------------
  


# reformulate_query substitutes the 'AND', 'OR' and 'NOT' operator words with the 
# corresponding boolean operators '&&', '||' and '!', respectively,  adds '&&' between
# words when they aren't separated by an operator, and additionally performs checks 
# on the entered query data. It takes a reference to a list as its sole argument. 

sub reformulate_query
{
    
    my $array_ref = shift();
    my $in_quotes = 0;
    my $last_non_empty = 0;
    my $i = 0;

    # First of all, were we passed an empty array?
    
    return unless (@$array_ref);
    
    # Loop through the array, performing various transformations:
    
    for  ($i = 0; $i < @$array_ref; $i++)
    {
	
	# Support phrases using '"':
	
	# Handle words beginning with double quotes:
	
	if ($array_ref->[$i] =~ /^\"(.+)/)
	{
	    # Special case: if the user, for whatever reason, enters a single word
	    # between double quotes, we remove them:

	    if ($array_ref->[$i] =~ /^\"(.+)\"$/)
	    {
		$array_ref->[$i] = $1;
		next;
	    }
		  
	    # If we are passed a word starting with double quotes and $in_quotes is already set, abort:
	    if ($in_quotes)
	    {
		print h3('Invalid query. You may not use double quotes in the quoted text.'), end_html;
		exit(1);
	    }
	    # Lose the double quotes and flag that this word began a quoted phrase:
	    $array_ref->[$i] = $1;
	    $in_quotes = 1;
	    next;
	}
	
	# Handle words ending in double quotes:
	
	if ($array_ref->[$i] =~ /(.+)\"$/)
	{
	    # If we are passed a word ending in double quotes but $in_quotes isn't set, abort:
	    unless ($in_quotes)
	    {
		print h3('Invalid query. Your query contains unmatched closing double quotes.'), end_html;
		exit(1);
	    }
	    # Append this word to the previous expression (at $i-1), putting a space-matching
	    # pattern between them. Then remove this item - not forgetting to rewind the counter 
	    # one unit so that we don't "jump" over the next list item after the "for" loop increments
	    # it - and signal that we just reached the end of a quoted phrase:
	    $array_ref->[$i - 1] .= '\s+' . $1;
	    splice @$array_ref, $i, 1;
	    $i--;
	    $in_quotes = 0;
	    next;
	}
	
	# Abort if we find double quotes in the middle of a word:
	
	if ($array_ref->[$i] =~ /.+\".+/)
	{
	    print h3('Invalid query. You may not insert double quotes in the middle of a word.'), end_html;
	    exit(1);
	}
	
	# Handle words between double quotes: just append the word to the 
	# phrase which is being stored at $i - 1 (separating them with a
	# space-matching pattern) and remove this item, also decrementing
	# the counter for the above-mentioned reason:
	if ($in_quotes) 
	{
	    $array_ref->[$i - 1] .= '\s+' . $array_ref->[$i];
	    splice @$array_ref, $i, 1;
	    $i--;
	    next;
	}
	
	# From this point on, we are handling a non-quoted word,
	# which might be an operator:
	
	# Convert "and", "or" and "not" to their symbols:
	
	if ($array_ref->[$i] =~ /\band\b/io) 
	{
	    $array_ref->[$i] = '&&';
	}
	
	if ($array_ref->[$i] =~ /\bor\b/io) 
	{
	    $array_ref->[$i] = '||';
	}
	
	# If the user enters the word 'NOT', we might have to translate it to '&& !', because 
	# most people (probably) just write 'a NOT b', rather than 'a AND NOT b'; and we need the 
	# latter formulation so that our matching algorithm - which relies on eval() - can correctly 
	# process this condition. If the '!' operator was indeed already preceded by either '&&' or
	# '||', don't do anything:
	
	if ($array_ref->[$i] =~ /\bnot\b/io && !is_op($array_ref->[$i-1]))
	{
	    $array_ref->[$i] = '&& !';
	}
	
	# Braces can be used for defining the correct evaluation order of the entered
	# match conditions, overriding the default operator precedence. If they are 
	# present, we must separate them from words:
	
	if ($array_ref->[$i] =~ /^\((.+)/o) 
	{
	    splice @$array_ref, $i, 1, '(', $1;
	}
	
	if ($array_ref->[$i] =~ /(.+)\)$/o) 
	{
	    # In this case, so that the word is processed,
	    # we must decrement $i:
	    splice @$array_ref, $i, 1, $1, ')';
	    $i--;
	}
	
    }
    
    # We must still loop through the loop to add "AND" between words which aren't
    # connected by an operator: "A B" should translate into "A AND B".

    for ($i = 1; $i < @$array_ref; $i++)
    {
	if ( !(is_op($array_ref->[$i])) && !($array_ref->[$i] =~ /[\(\)]/) &&
	    !(is_op($array_ref->[$i - 1])) && !($array_ref->[$i - 1] =~ /[\(\)]/) )
	{
	    # Just insert an 'AND' between those two words:
	    splice @$array_ref, $i, 0, '&&';
	}
    }
    
    
    # Try to detect malformed queries: --------------------------
    # I would prefer to perform these checks at the beginning of the subroutine,
    # but this is much easier to do after the list has been processed by the code
    # above. 
    
    # If $in_quotes is still set, this means we are lacking the closing double quotes:
    
    if ($in_quotes)
    {
	print h3('Invalid query. You forgot the closing double quotes.'), end_html;
	exit(1);
    }
    
    # Abort if the query text either begins or ends with an operator:
    
    if (is_op($array_ref->[0]) || is_op($array_ref->[@$array_ref - 1]))
    {
	print h3('Invalid query. You may not begin or terminate a query with an operator.'), end_html;
	exit(1);
    }
    
    # Finally, concatenate all tokens into a string, while replacing all words
    # with '1', and 'eval()' it. If eval returns a undefined value, it means the 
    # query isn't valid:
    
    my $test_string = '';

    my $token;
    
    foreach $token (@$array_ref) 
    {
	if ($token eq '(' || $token eq ')' || is_op($token))
	{
	    $test_string .= ' ' . $token;
	}
	else 
	{
	    $test_string .= ' ' . 1;
	}
    }
    unless (defined(eval($test_string)))
    {
	print h3('Invalid query. Please read the search instructions.'), end_html;
	exit(1);
    }
    
    # All seems to be fine; quit. The array is now in the correct form.
    return;    
}

# -------------------------




# match_query, used to check for matches... :


sub match_query
{
    # $text contains the sentence / paragraph ("etext_atom") which might
    # match the query's conditions:
    my $text = shift();
    
    # @query holds the query's expression as a list of individual words and
    # operators (e.g., "A and B" -> ('A', '&&', 'B')):
    my @query = @ARG;
    
    # If we were passed an empty array, then return TRUE, since this means the user didn't
    # specify any search word in this field, and therefore we have got nothing to test $text
    # against:
    
    return 1 unless (@query);
    
    # This loop puts together a string which contains the elements found in @query, with the
    # exception that all the words which aren't boolean operators have been replaced with 
    # either a 1, if they were found in $text, or a 0, if they weren't:
    
    my $eval_expression = '';
    
    my $token;
    
    foreach $token (@query)
    {
	# If this token is an operator (or a brace), append it to $eval_expression and jump
	# to the next iteration:
	if ($token eq '(' || $token eq ')' || is_op($token))
	{
	    $eval_expression .= ' ' . $token;
	    next;
	}
	# If this token isn't an operator and can be found in $text, append a 1 (TRUE)
	# to the string:
	elsif ($text =~ /\b$token\b/i)
	{
	    $eval_expression .= ' 1';
	}
	# If this token isn't an operator and hasn't been found in $text, append a 0
	# (FALSE) to $eval_expression:
	else 
	{
	    $eval_expression .= ' 0';
	}
    }
    
    # At this point, $eval_expression looks something like, e.g., ' ( 1 && 1 )' or
    # ' ( 1 || 1 && 0 )'. By calling eval() on such an expression, we can have it
    # compute its value for us: if it returns 1, it means that $text obeys the search
    # criteria:
    
    if (eval("$eval_expression"))
    {
	return 1;
    }
    else
    {
	return 0;
    }
    
}

# --------------------


# This function returns the contents of the specified file.
# In case of failure, we return undef.

sub get_local_file
{
    my $file_path = shift();
    
    # If the argument begins with 'file:/', remove it so
    # that we have an absolute path:
    if ($file_path =~ m{^file:/.+}io)
    {
	$file_path = substr($file_path, length("file:/"));
    }
    
    open FILE, $file_path or return undef;
    undef $INPUT_RECORD_SEPARATOR;
    my $contents = <FILE>;
    close FILE;
    
    return $contents;
}


# ------------------------------------------------------


# This function returns the contents of the specified directory.
# In case of failure, we return an empty list.

sub get_local_dir
{
    my $dir = shift();
    opendir DIR, $dir or return ();
    
    return readdir DIR;
}
    



# -------------------------------------------------------

# This function returns the contents of the specified file.
# In case of failure, we return undef.

sub get_remote_file
{
    my $url = shift();
    
    # Retrieve the file using function get() from LWP::Simple:
    return get($url);
}
    


# ----------------------------------------------------------


# It is more complex than its "sister" get_local_dir because HTTP and
# FTP servers, when asked for a directory listing, return more information
# than we need. In case of failure, we return an empty list.

sub get_remote_dir
{
    my $url = shift();
    
    # Store the server's reply in the scalar 
    # $directory_listing:
    my $directory_listing = get($url);
    
    # If it failed, quit:
    return () unless (defined($directory_listing));
    
    # Begin by splitting the server's reply into lines:
    
    my @dir_lines = split /\n/, $directory_listing;
    
    # The "treatment" applied to @dir_lines to extract the filenames
    # depends on the protocol which was used to retrieve it - HTTP or FTP:
    
    # If @dir_lines contains a directory listing embedded in HTML, we (1)
    # replace the lines which describe a file with just the file's name and
    # (2) remove any lines which don't refer to a file in that directory from
    # the array:
    
    my $i;
    
    if ($url =~ m{^http://.+})
    {
	
	for ($i = 0; $i < @dir_lines; $i++)
	{
	    # I arrived at this regex by examining the ouput produced by
	    # requesting a directory listing from the HTTP server at
	    # ibiblio.org. If this fails with other servers, please let
	    # me know. It works by extracting the filename from both
	    # the link reference as well as its caption:
	    if ($dir_lines[$i] =~ m{^<IMG SRC=[^>]+>\s+<A HREF=\"([^\"]+)\">\1</A>.+})
	    {
		# This is a line which describes a file in this dir, so replace
		# it with the filename stored in $1:
		$dir_lines[$i] = $1;
	    }
	    else
	    {
		# If this line doesn't refer to a file in this dir, remove it and
		# rewind the counter one unit:
		splice @dir_lines, $i, 1;
		$i--
	    }
	}
    }
    else # Handle directory listings obtained from FTP servers (once again, the regex
      # I am using is derived from my examination of the output from the FTP server 
      # at ibiblio.org):
    {
	
	for ($i = 0; $i < @dir_lines; $i++)
	{
	    if ($dir_lines[$i] =~ /.+\s+(\w+\.\w+)/)
	    {
	    # This line contains a filename:
		$dir_lines[$i] = $1;
	    }
	    else
	    {
		# This one doesn't, so remove it from the array and decrement
		# the iteration counter:
		splice @dir_lines, $i, 1;
		$i--;
	    }
	}
    }
    # In any case, we now have the names of all files in this directory
    # listed in $dir_lines. The only thing left to do is to return that
    # list to the calling function:
    
    return @dir_lines;
}

# ------------------------------------------------


# This subroutine accepts a reference to a headerless etext as its sole
# argument and then decomposes it into individual sentences, returning them
# in an array:

sub separate_sentences
{
    my $etext_ref = shift();
    
    my @sentences;
    
    my $i;
    
    # Split the etext into sentences, not forgetting to store the punctuation
    # marks in the array for now (through the use of braces in the pattern):
    
    @sentences = split /([.!?]+)/, $$etext_ref;

    # Append the punctuation marks to the sentence which precedes them, and
    # remove them from the array (also rewinding the counter):
    
    for ($i = 1; $i < @sentences; $i++)
    {
	if ($sentences[$i] =~ /[.!?]+/)
	{
	    $sentences[$i - 1] .= $MATCH;
	    splice @sentences, $i, 1;
	    $i--;
	}
	# Handle unmatched double quotes (by preceding the sentence with - or
	# appending to it - "[...]"):
	if ($sentences[$i] =~ /^\s*\"[^\"]+$/)
	{
	    $sentences[$i] .= " [...]\"";
	}
	elsif ($sentences[$i] =~ /^[^\"]+\"\s*[.?!]+/)
	{
	    substr ($sentences[$i], 0, 0, "\"[...] ");
	}
    }
    
    return @sentences;
}

# ------------------------------------------



# Being passed a line from the index file, this function returns the individual 
# items. In case of failure, it returns an empty list.
    

sub extract_etext_data
{
    my $line = shift();
    
    my @fields = split /\t/, $line;
    
    # Did we process the line correctly?
    
    # The indexes used come from the HOWTO.TXT file describing
    # the format of the index file used at the time of writing:
    
    return () unless ($fields[1] ne '' && 
		      $fields[4] ne '' && 
		      $fields[@fields - 1 - 2] ne '' && 
		      $fields[@fields - 1 - 3] ne '');
		      
    # Remove file name extension from the file name:
    
    $fields[@fields - 1 - 2] =~ s/([^\.]+).+/$1/;
    
    # Return author, title, generic filename and directory:
    
    return ($fields[1], $fields[4], $fields[@fields - 1 - 2], $fields[@fields - 1 - 3]);
}

# ------------------------------------------------


# This subroutine removes the header with which all PG etexts begin, so that 
# we just scan the book itself, not the "small print!".

# It is very simple: it searches for a line containing a string ressembling
# "END THE SMALL PRINT" and cuts all text from the beginning of the file to
# the first empty line which follows it. It takes as its sole argument a 
# reference to a PG etext. (Since some etexts are pretty long, I suppose that 
# it is worth using a reference.)

sub remove_header
{
    my $etext_ref = shift();
    
    # Marks the end of small print:
    my $end_sm_pr = -1;
    
    # Marks the end of the paragraph in which we found $end_sm_pr:
    my $end_paragraph = -1;
    
    # These strings are based on my examination of the different "small print!"
    # versions found in the PG archive. If I am missing something, please
    # let me know:
    
    
    my @end_sm_pr_strings = ("END THE SMALL PRINT! FOR PUBLIC DOMAIN ETEXTS",
			       "END*THE SMALL PRINT! FOR PUBLIC DOMAIN ETEXTS",
			       "*SMALL PRINT! Ver.06.28.92* Zen and the Art of the Internet*END*");
    my $string;
    
    # If we find one of the strings above, we exit the loop:
    foreach $string (@end_sm_pr_strings)
    {
	$end_sm_pr = index $$etext_ref, $string;
	last if ($end_sm_pr != -1);
    }
    
    # If we didn't find the any of the strings above, we do nothing and return. Let's
    # just hope that we don't show the user any matches from the header, if it is present:
    
    return unless ($end_sm_pr != -1);
    
    
    # At this point, we must look for the end of the paragraph which contains the "end
    # the small print!" passage:
    
    $end_paragraph = index $$etext_ref, "\n\n", $end_sm_pr;
    
    # If we couldn't find the end of that paragraph, we just cut everything up to
    # the end of line which follows $end_sm_pr, if we can find it:
    
    if ($end_paragraph == -1)
    {
	my $eol = index $$etext_ref, "\n", $end_sm_pr;
	if ($eol != -1)
	{
	    substr($$etext_ref, 0, $eol) = '';
	}
	else # if, for some strange motive we didn't find the EOL:
	{
	    substr($$etext_ref, 0, $end_sm_pr) = '';
	}
    }
    else
    {
	# Otherwise, if we did find the end of that paragraph, remove everything up
	# to $end_paragraph:
	substr($$etext_ref, 0, $end_paragraph) = '';
    }
    
    # We are done:
    
    return;
}
    
 

# -----------------------------------------------------


# Very simple: 
sub separate_paragraphs
{
    
    my $etext_ref = shift();
    
    return split /\n\s*\n/, $$etext_ref;
    
}


# -------------------------------------------------------


sub decompress_zipped_etext
{
    # Variable which holds the contents of the retrieved zipped etext:
    
    my $compressed_contents_ref = shift();
    
    # Variable which will hold the decompressed etext:
    
    my $decompressed_contents;
    
    # Etext name (needed to access the compressed file inside the zip archive):
    
    my $etext_name = shift();
    
    
    # We need to "pretend" that we have a file handle the zip file and then
    # write it unscrambled to a second file handle; we do this through the 
    # use of the module IO::Scalar:
    
    my $scalar_handle_1 = IO::Scalar->new($compressed_contents_ref);
    my $scalar_handle_2 = IO::Scalar->new(\$decompressed_contents);
    
    # Create a zip archive:
    
    my $zip_archive = Archive::Zip->new();
    
    # Fill it with the data found in $compressed_contents (via $scalar_handle_1):
    
    return undef if ($zip_archive->readFromFileHandle($scalar_handle_1) != AZ_OK);
    
    # Store the zipped file with the name $etext_name . "txt" in the scalar 
    # $compressed_file:
    
    my $compressed_file = $zip_archive->memberNamed($etext_name . ".txt");
    
    return undef unless (defined($compressed_file));
    
    # Decompress $compressed_file into the variable $decompressed_contents (via 
    # $scalar_handle_2):
    
    return undef if ($compressed_file->extractToFileHandle($scalar_handle_2) != AZ_OK);
    

    $scalar_handle_1->close();
    $scalar_handle_2->close();
    
    # Return decompressed etext! 
    
    return $decompressed_contents;
    
}




# ---------------- End of helper subroutines --------------------



# -------------------------- Beginning -----------------------------

# Begin outputting page:

print header, 
  start_html(-title => 'Search results');

# First, by looking at $pg_root_dir, determine whether remote or local
# access should be used and point $getfile and $getdir to the correct 
# functions.
# $getfile and $getdir are references to the file/directory retrieval 
# functions we will use. If a local PG mirror was specified, they will
# be set to \&get_local_file and \&get_local_dir, respectively; if that 
# is not the case, then they will  point to &get_remote_file and 
# &get_remote_dir.

my ($getfile, $getdir);

if ($pg_root_dir =~ m{^file://.+}i ||
    $pg_root_dir =~ m{^/.+})
{
    $getfile = \&get_local_file;
    $getdir = \&get_local_dir;
}
elsif ($pg_root_dir =~ m{^(ftp://|http://).+}i)
{
    $getfile = \&get_remote_file;
    $getdir = \&get_remote_dir;
}
else 
{
    print h3("Error processing the configuration variable. Please inform the system administrator."), 
      end_html;
    exit(1);
}

# Check whether the provided path/URL ends with a slash, and, if
# it doesn't, append one:

$pg_root_dir .= '/' unless ($pg_root_dir =~ m{.+/&});
      

# Initialize some variables:

# Array of the lines forming the index file:
my @index_lines; 

# Used to store the values entered in the form by the user:
my ($author_string, $title_string, $words_string, $search_type);

# Arrays in which we store the individual words after decomposing the three strings declared above:
my (@author_names, @title_words, @words); 


my $i = 0; # generic counter

# Open index file, quit if it fails:

@index_lines = split /\n/, &$getfile($pg_root_dir . 'pgdb.txt');


unless (@index_lines)
{
    print h3('Unable to open index file. Please inform the system administrator.'), 
      end_html;
    exit(1);
}

# Get the parameters from the CGI form:

$author_string = param('author');
$title_string  = param('title');
$words_string  = param('words');
$search_type   = param('type');

# Decompose the strings containing the parameters into arrays:

@author_names = split /\s+/, $author_string;
@title_words  = split /\s+/, $title_string;
@words        = split /\s+/, $ words_string;

# Test parameters, abort if something's missing:

unless (@words)
{
    print h3('You must enter at least one word in the "Words:" field.'), end_html;
    exit(1);
}

if ($search_type ne 'paragraph' && $search_type ne 'sentence') 
{
    print h3('You must select a search type; choose either "paragraph" or "sentence".'),
      end_html;
    exit(1);
}

unless (@author_names || @title_words)
{
    print h3('You must enter at least either an author name or a title word.'), end_html;
    exit(1);
}

# Rewrite operators and perform various checks on the entered data:

reformulate_query(\@author_names);
reformulate_query(\@title_words);
reformulate_query(\@words);

# Recall the user of what she inserted in the form:

print h3("Results of your query"), 
  var("author: $author_string"), p,
  var("title: $title_string"), p,
  var("words: $words_string"), p,
  var("type: $search_type"), p,
  hr;


# Loop through the index file and find works which match the author/title:

# Initialize some more variables:

# Holds each index line at a time:

my $line;

# Keeps track of how many etext entries in the index file match the specified author/title

my $etext_matches = 0; 

# Used to store data about each etext entry while processing the index file:

my ($etext_author, $etext_title, $etext_file, $etext_dir);

foreach $line (@index_lines)
{
    # Did we correctly process this line? If we didn't, jump to the next iteration:

    next unless (($etext_author, $etext_title, $etext_file, $etext_dir) = extract_etext_data($line));
    
    # Check whether the author and title match the entered keywords; if they don't, 
    # then end this iteration:
    
    next unless (match_query($etext_author, @author_names) &&
		 match_query($etext_title, @title_words));
    
    # If we get to this point, the etext entry matches our query (as far as author and title are 
    # concerned) and we need to grep it:
    
    # First, increment $etext_matches and find out which is the latest version of this etext:
    
    $etext_matches++;
    
    
    # Retrieve the listing using getdir and store its lines in an array:
    my @year_files = &$getdir($pg_root_dir . $etext_dir);
    
    
    # Quit if it fails:
    unless (@year_files)
    {
	print h3("Unable to open directory $etext_dir. Please inform the system administrator."), 
	  end_html;
	exit(1);
    }
    
    
    # Scan all files which match the etext filename pattern, and choose the text version, if available.
    # We support retrieving zipped etexts, too, but when a text version is available we
    # prefer it (this is done automatically, since directory listings are lexicographically
    # ordered and "etextNN.txt" comes before "etextNN.zip"):
    
    my $format = '';
    my $filename;
    
    foreach $filename (@year_files)
    {
	if ($filename =~ /$etext_file\.(\w{3})$/)
	{
	    $format = $1;
	    last;
	}
    }
    
    if ($format eq '')
    {
	print h3("Unable to find a version of file ${etext_file}.* in dir $etext_dir. Please inform the system administrator."),
	  end_html;
	exit(1);
    }
    
    # Retrieve etext file and store it in $etext:
    
    my $etext_filename = $etext_file . ".$format";
    
    my $complete_etext_path = $pg_root_dir . $etext_dir . "/" . $etext_filename;
    
    my $etext = &$getfile($complete_etext_path);
    
    # Quit if it fails:
    unless (defined($etext))
    {
	print h3("Unable to open file $complete_etext_path. Please inform the system administrator."), 
	  end_html;
	exit(1);
    }
    
    # If the file is zipped, uncompress it:

    $etext = decompress_zipped_etext(\$etext, $etext_file) if ($format eq 'zip');
    
    # Check if decompress_zipped_etext exited with success:
    
    unless (defined($etext))
    {
	print h3("Error while decompressing zipped etext $etext_filename. Please inform the system administrator"),
	  end_html;
	exit(1);
    }
    
    # PG's etexts all have a header which we must remove before scanning
    # the file. remove_header does precisely that (I hope):
    
    remove_header(\$etext);
    
    # Now, according to the chosen $search_type, we must decompose $etext into
    # either a list of paragraphs ("logical lines") or a list of sentences:
    
    my @etext_atoms;
    
    if ($search_type eq 'paragraph')
    {
	@etext_atoms = separate_paragraphs(\$etext);
    }
    else # if ($search_type eq 'sentence'):
    {
	@etext_atoms = separate_sentences(\$etext);
    }
    
    # $etext_atom will hold either a sentence or a paragraph, depending on the
    # chosen search type:
    
    my $etext_atom;
    
    # Loop through the file, seeing if each $etext_atom matches the specified
    # words, and store the relevant passages in the array @results, so that we
    # can print them all _preceded_ by a short text naming the book they were 
    # found in together with its author:
    
    my @results;
    
    # Keep a record of the number of matches:
    
    my $matches = 0;
	
    foreach $etext_atom (@etext_atoms)
    {
	if (match_query($etext_atom, @words))
	{
	    # If this $etext_atom matches @words, we increment the counter of
	    # matches and append this $etext_atom to the array @results:
	    
	    $matches++;
	    push @results, "$matches- $etext_atom", p;
	}
    }
    
    # Output results depending on whether there were any matches:
    
    if ($matches)
    {
	# If there were any matches, prefix the relevant sentences/paragraphs
	# with a short header informing the user where they were found:
	
	unshift @results, h3("The following" , $matches > 1 ? "$matches ${search_type}s" : "$search_type",
			     "taken from", i("$etext_title,"), "a book by",
			     i("$etext_author,"), $matches > 1 ? "match" : "matches", "your query:"), p;
	
	# Print out the results:
	print @results;
    }
    else # If there weren't any matches, say so:
    {
	    print h3("The etext", i("$etext_title,"), "written by",
		     i("$etext_author,"), "was searched but there weren't any matches."), p;
    }
    
    # Print an horizontal ruler to separate this etext's results from the following:
    print hr;    
}

# After looping through the index, if we couldn't find any relevant entries in 
# the index file, say so:

unless ($etext_matches)
{
    print h3("Sorry, no etexts were found which matched your query:"), p;
    
}

# End HTML:
print end_html;



# ---------------------------------------- End -----------------------


# Copyright 2001 Manuel Arriaga
